from abc import ABC


class Constants(ABC):
    # pylint: disable = invalid-name
    # defaults to ultrastar and see-level
    AREA: float = 0.058556  # m ^ 2
    I_xx: float = 0.001219  # kg * m ^ 2
    I_zz: float = 0.002352  # kg * m ^ 2
    MASS: float = 0.175  # kg
    RH0: float = 1.225  # air density, kg / m ^ 3
    GRAVITY: float = 9.81  # m / s ^ 2


class FrispyUltrastarConstants(Constants):
    AREA: float = 0.058556  # m ^ 2
    I_xx: float = 0.001219  # kg * m ^ 2
    I_zz: float = 0.002352  # kg * m ^ 2
    MASS: float = 0.175  # kg
